#include "module.h"

#include "rose.h"

#include "design.h"

#include "../tools/misc.h"
#include "../tools/class_is_derived_from.h"

#include <algorithm>

risc::Module::Module(SgClassDefinition *ast_node):
  Class(ast_node)
{
  initialize_variables();
  initialize_functions();
}

risc::Module::~Module()
{
  for(VariableVectorIter iter = variables_.begin(); iter != variables_.end();
      iter++) {
    delete *iter;
  }

  for(ModuleVectorIter iter = sub_modules_defs_.begin();
      iter != sub_modules_defs_.end(); iter++ ) {
    delete *iter;
  }

  for(ModuleInstanceVectorIter iter = sub_modules_.begin();
      iter != sub_modules_.end(); iter++) {
    delete *iter;
  }

  for(PrimitiveChannelInstanceVectorIter iter = sub_prim_channel_.begin();
      iter != sub_prim_channel_.end(); iter++) {
    delete *iter;
  }

  for(HierarchicalChannelInstanceVectorIter iter = sub_hier_channel_.begin();
      iter != sub_hier_channel_.end(); iter++) {
    delete *iter;
  }

  for(PrimitiveChannelVectorIter
      iter = sub_prim_channel_defs_.begin();
      iter != sub_prim_channel_defs_.end(); iter++) {
    delete *iter;
  }

  for(HierarchicalChannelVectorIter
      iter = sub_hier_channel_defs_.begin();
      iter != sub_hier_channel_defs_.end(); iter++) {
    delete *iter;
  }

  for(ThreadVectorIter iter = threads_.begin(); iter != threads_.end();
      iter++) {
    delete *iter;
  }

  for(CThreadVectorIter iter = cthreads_.begin(); iter != cthreads_.end();
      iter++) {
    delete *iter;
  }

  for(MethodVectorIter iter = methods_.begin(); iter != methods_.end();
      iter++) {
    delete *iter;
  }

  for(FunctionVectorIter iter = functions_.begin(); iter != functions_.end();
      iter++) {
    delete *iter;
  }

  for(InPortVectorIter iter = inports_.begin(); iter != inports_.end();
      iter++) {
    delete *iter;
  }

  for(InOutPortVectorIter iter = inoutports_.begin();
      iter != inoutports_.end(); iter++) {
    delete *iter;
  }

  for(OutPortVectorIter iter = outports_.begin(); iter != outports_.end();
      iter++) {
    delete *iter;
  }

  for(EventVectorIter iter = events_.begin(); iter != events_.end();
      iter++) {
    delete *iter;
  }
}

void risc::Module::initialize_variables()
{
  // This loop iterates all members of the class. A member can be a
  // constructor, typedef, function, and so on. For this case we have to
  // distinguish the different types
  for(SgDeclarationStatementPtrList::iterator
      iter  = get_ast_node()->get_members().begin();
      iter != get_ast_node()->get_members().end(); iter++) {

    SgVariableDeclaration* vd = isSgVariableDeclaration(*iter);

    // We are only interested in variable declarations
    if(!vd)
      continue;

    SgVariableDefinition *var_def = vd->get_definition();

    // We check if it is a SystemC element
    //
    // A sc_module instance has two significant attributes:
    //  a) the instance is a data structure (class)
    //  b) the class is derived from sc_module
    //
    // A port
    //  a) derived from sc_port
    if(risc::tools::is_data_structure(var_def->get_type())) {

      SgClassType* ct = isSgClassType(var_def->get_type());

      SgClassDeclaration *declaration
        = isSgClassDeclaration(ct->get_declaration());

      if(declaration->isForward()) {
        declaration =
          isSgClassDeclaration(declaration->get_definingDeclaration());
      }

      SgClassDefinition *definition = declaration->get_definition();

      //Okay, at this point we know it is a C++ data structure.

      //case of a module
      if(risc::tools::is_derived_from_sc_module(definition) ) {

        // At this point, we have the SgClassDefinition of the corresponding
        // class member variable in the module. In the next step, we have to
        // create a Module instance and insert into the design

        Module *md =
          GlobalDesign::design_static->find_module_definition(definition);

        sub_modules_.push_back(new ModuleInstance(var_def, md));
        continue;
      }

      // case of a port
      if(risc::tools::is_derived_from_sc_port_base(definition) ) {


        SgTemplateInstantiationDefn* templ_def =
          isSgTemplateInstantiationDefn(definition);
        if(templ_def) {

          // case of an in_port
          if(risc::tools::is_sc_in_port(templ_def)){

            inports_.push_back(new InPort(var_def));
            continue;
          }

          // case of an out_port
          if(risc::tools::is_sc_out_port(templ_def)){

            outports_.push_back(new OutPort(var_def));
            continue;
          }

          // case of an in_port
          if(risc::tools::is_sc_inout_port(templ_def)){

            inoutports_.push_back(new InOutPort(var_def));
            continue;
          }
        }

        // in any other case we can not define it. So it must be an inout port
        inoutports_.push_back(new InOutPort(var_def));
        continue;
      }

      // case of a event
      if(risc::tools::is_sc_event(definition)) {

        events_.push_back(new Event(var_def));
        continue;
      }

      // case of a primitive channel
      if(risc::tools::is_derived_from_sc_prim_channel(definition)) {

        sub_prim_channel_.push_back(new PrimitiveChannelInstance(var_def,NULL));
        continue;
      }

      // case of a channel (this case represents a hierarchical channel)
      if(risc::tools::is_derived_from_sc_channel(definition)) {

        // FIXME
        // A hierarchical has to have a user definition!
        // This is only for test cases
        sub_hier_channel_.push_back(
            new HierarchicalChannelInstance(var_def,NULL));
        continue;
      }
    }

    // We check if it is an int, double, float, unsigned, ...
    if(risc::tools::is_builtin_type(var_def->get_type()) ||
       risc::tools::is_data_structure(var_def->get_type())) {

      variables_.push_back(new Variable(var_def));
    }

    if(SageInterface::isReferenceType(var_def->get_type())) {

      variables_.push_back(new Variable(var_def));
    }
  }
}


void risc::Module::initialize_functions()
{
  // Traversing for functions and getting all associated SystemC semantics is a
  // little bit tricky. We have to distinguish between a function in the
  // classic C++ style and a SystemC process which can have a static
  // sensitivity list. We can obtain these information in the ctor of the
  // individual module. This is the concept:
  //
  // (1) We traverse for all function definitions in the class definition
  //     and put them in a temporary container
  // (2) We have to analyze the ctor of the module and figure out which
  //     functions are registrated as Thread, CThread, and Method.
  // (3) We pick these function from the temporary container and put them
  //     into the corresponding process lists for threads, cthreads, and
  //     methods
  // (4) We analyze the following static sensitivity list (detials later)
  //     and associate the sensitivity list with the SystemC processes
  // (5) Finally, only 'pure' C++ function are in the temporary container

  // get ctor of the class (module)
  SgMemberFunctionDeclaration *ctor =
    SageInterface::getDefaultConstructor(get_ast_node()->get_declaration());

  // Find all functions in the class defintion
  Rose_STL_Container<SgNode*> function_definitions
    = NodeQuery::querySubTree(get_ast_node(), V_SgFunctionDefinition);

  // The constructor is in the function_definitions container because rose
  // treats a constructor as a special kind of a function. So, we have to
  // search for the ctor and delete it.
  function_definitions.erase(
    std::remove(function_definitions.begin(), function_definitions.end(),
      ctor->get_definition()),
    function_definitions.end());

  // Now, we will decide if each individal function has a specific SystemC
  // semantic (like a THREAD, ...). Please consider how a SystemC module
  // constructor is transformed after the preprocessing below:

  /* INPUT of a SystemC ctor before preprocessing
   * Foo(sc_module_name mn):
   *   sc_module(mn)
   * {
   *   SC_THREAD(my_thread1);
   *
   *   SC_THREAD(my_thread2);
   *   sensitive << e1 << e2;
   * }
   *
   * OUTPUT of a SystemC ctor after preprocessing
   * inline Foo(class sc_module_name mn): sc_module(mn) {
   *   {
   *     class sc_process_handle my_thread1_handle =
   *       sc_core::sc_get_curr_simcontext()->create_thread_process(
   *         "my_thread1", false, &Foo::my_thread1, (this), 0);
   *     (this)->sensitive << my_thread1_handle;
   *     (this)->sensitive_pos << my_thread1_handle;
   *     (this)->sensitive_neg << my_thread1_handle;
   *   };
   *   {
   *     class sc_process_handle my_thread2_handle =
   *       sc_core::sc_get_curr_simcontext()->create_thread_process(
   *           "my_thread2", false, &Foo::my_thread2, (this), 0);
   *   (this)->sensitive << my_thread2_handle;
   *   (this)->sensitive_pos << my_thread2_handle;
   *   (this)->sensitive_neg << my_thread2_handle;
   *   };
   *   ((this)->sensitive << ((this)->e1)) << ((this)->e2);
   * }
   */
  // This example shows that each process registration is transformed in a
  // dedicated basic block. The sensitivity list is immediately after the
  // basic block.

  // First, we search for basic blocks in the ctor. In the basic block we
  // might expect the registration of the function to the sysc kernel.
  // For this reason, we check if the specific sub basic block in the ctor
  // includes a function call to create_XXX_process. If this is the case, we
  // are dealing with a function registration.
  // Afterwards, we have to search for event registrations. An event
  // registration happens immeditly after the basic block. So, we search for
  for(SgStatementPtrList::iterator
      iter  = ctor->get_definition()->get_body()->get_statements().begin();
      iter != ctor->get_definition()->get_body()->get_statements().end();
      iter++) {

    SgBasicBlock *bb = isSgBasicBlock(*iter);

    if(bb) {

      Rose_STL_Container<SgNode*> function_calls
        = NodeQuery::querySubTree(bb, V_SgFunctionCallExp);

      bool has_potential_sensitivity_list = false;

      for(Rose_STL_Container<SgNode*>::iterator
          iter_call = function_calls.begin();
          iter_call != function_calls.end(); iter_call++) {

        SgFunctionCallExp *call_exp = isSgFunctionCallExp(*iter_call);
        SgFunctionDeclaration *fd = call_exp->getAssociatedFunctionDeclaration();

        SgFunctionDefinition *process_def = NULL;
        Thread* new_thread = NULL;
        Method* new_method = NULL;

        if(fd->get_name() == "create_thread_process") {

          SgExprListExp* expr_list = call_exp->get_args();

          process_def =
              SageInterface::findFunctionDeclaration(
                SageInterface::getEnclosingClassDefinition(
                  expr_list),
                  isSgStringVal(((expr_list->get_expressions())[0]))->get_value(),
                  NULL,
                  true)->get_definition();

          has_potential_sensitivity_list = true;
          new_thread = new Thread(process_def);
          threads_.push_back(new_thread);
        }

        if(fd->get_name() == "create_cthread_process") {

          SgExprListExp* expr_list = call_exp->get_args();

          process_def =
              SageInterface::findFunctionDeclaration(
                SageInterface::getEnclosingClassDefinition(
                  expr_list),
                  isSgStringVal(((expr_list->get_expressions())[0]))->get_value(),
                  NULL,
                  true)->get_definition();

          cthreads_.push_back(new CThread(process_def));
        }

        if(fd->get_name() == "create_method_process") {

          SgExprListExp* expr_list = call_exp->get_args();

          process_def =
              SageInterface::findFunctionDeclaration(
                SageInterface::getEnclosingClassDefinition(
                  expr_list),
                  isSgStringVal(((expr_list->get_expressions())[0]))->get_value(),
                  NULL,
                  true)->get_definition();

          has_potential_sensitivity_list  = true;
          new_method = new Method(process_def);
          methods_.push_back(new_method);
        }

        // Delete the process in the temporary function list
        if(process_def) {
          function_definitions.erase(
            std::remove(function_definitions.begin(),
              function_definitions.end(), process_def),
            function_definitions.end());
        }

        // Now, we search for a static sensitivity list
        // The static sensitivity list should follow immediately after the
        // registration of the function. General idea: We analyze statement for
        // statement until the next statement is a basic block
        if(has_potential_sensitivity_list) {
          SgStatementPtrList::iterator previous_iter = iter;

          // get next statement in the constructor
          iter++;

          // In this loop, we iterate the constructor body statements until we
          // reach the end of the constructor body or the next statement is a
          // new basic block
          while(iter != ctor->get_definition()->get_body()->get_statements().end()) {

            if(isSgExprStatement(*iter)) {

              // Find all variable references in the current statement.
              // In the statement itself, we will look for references with the
              // type of an event
              Rose_STL_Container<SgNode*> events
                = NodeQuery::querySubTree(*iter, V_SgVarRefExp);

              for(Rose_STL_Container<SgNode*>::iterator
                  event_iter  = events.begin();
                  event_iter != events.end(); event_iter++) {

                SgVarRefExp* var_ref = isSgVarRefExp(*event_iter);

                // We are looking for all variable references with the type of
                // an event
                if(var_ref->get_type() ==
                    risc::get_design()->sc_event_->get_declaration()->get_type()) {

                  SgVariableDefinition* var_def = isSgVariableDeclaration(
                      var_ref->get_symbol()->get_declaration()->
                        get_declaration())->get_definition();

                  Event* event = find_event(var_def);

                  if(new_thread)
                    new_thread->static_sensitivity_list_.push_back(event);
                  else
                    new_method->static_sensitivity_list_.push_back(event);

//                  std::cout << SageInterface::get_name(process_def)
//                            << ",   xoxo "
//                            << find_event(var_def)->get_name()
//                            << std::endl;
                }
              }

              previous_iter = iter;
              iter++;

            } else {
              break;
            }
          }

          iter = previous_iter;

        }
      }
    }
  }


  // All other functions are classic C++ functions
  for(Rose_STL_Container<SgNode*>::iterator
      iter = function_definitions.begin();
      iter != function_definitions.end(); iter++) {

    functions_.push_back(new Function(isSgFunctionDefinition(*iter)));
  }
}


void risc::Module::print_module_definition()
{
  std::cout << "Definition of " << get_ast_type_name() << std::endl;

  // print module variables
  if(sub_modules_.size() != 0) {

    std::cout << " Submodules" << std::endl;
    for(ModuleInstanceVectorIter iter = sub_modules_.begin();
       iter != sub_modules_.end(); iter++) {

      std::cout << "  " << (*iter)->get_ast_type_name()
                << " " << (*iter)->get_name() << std::endl;
    }
  }

  // print builtin variables
  if(variables_.size() != 0) {

    std::cout << " Variables" << std::endl;
    for(VariableVectorIter iter = variables_.begin();
      iter != variables_.end(); iter++) {

      std::cout << "  " << (*iter)->get_ast_type_name()
                << " " << (*iter)->get_name() << std::endl;
    }
  }

  // print events
  if(events_.size() != 0) {

    std::cout << " Events" << std::endl;
    for(EventVectorIter iter = events_.begin();
        iter != events_.end(); iter++) {

      std::cout << "  " << (*iter)->get_ast_type_name()
                << " " << (*iter)->get_name() << std::endl;
    }
  }

  // print ports
  if( (inports_.size() != 0) ||
      (outports_.size() != 0) ||
      (inoutports_.size() != 0)) {

    std::cout << " Ports" << std::endl;
    for(InPortVectorIter iter = inports_.begin();
        iter != inports_.end(); iter++) {

      std::cout << "  In    " << (*iter)->get_ast_type_name()
                << " " << (*iter)->get_name() << std::endl;
    }

    for(OutPortVectorIter iter = outports_.begin();
        iter != outports_.end(); iter++) {

      std::cout << "  Out   " << (*iter)->get_ast_type_name()
                << " " << (*iter)->get_name() << std::endl;
    }

    for(InOutPortVectorIter iter = inoutports_.begin();
        iter != inoutports_.end(); iter++) {

      std::cout << "  InOut " << (*iter)->get_ast_type_name()
                << " " << (*iter)->get_name() << std::endl;
    }
  }

  // print primitive channels
  if(sub_prim_channel_.size() != 0) {
    std::cout << " Primitive Channels" << std::endl;

    for(PrimitiveChannelInstanceVectorIter iter = sub_prim_channel_.begin();
        iter != sub_prim_channel_.end(); iter++) {

      std::cout << "  " << (*iter)->get_ast_type_name()
                << " " << (*iter)->get_name() << std::endl;
    }
  }

  // print hierarchical channels
  if(sub_hier_channel_.size() != 0) {

    std::cout << " Hierarchical Channels" << std::endl;
    for(HierarchicalChannelInstanceVectorIter iter = sub_hier_channel_.begin();
        iter != sub_hier_channel_.end(); iter++) {

      std::cout << "  " << (*iter)->get_ast_type_name()
                << " " << (*iter)->get_name() << std::endl;
    }
  }

  // print member functions
  if(functions_.size() != 0) {

    std::cout << " Member Functions (type, name)" << std::endl;
    for(FunctionVectorIter iter = functions_.begin();
        iter != functions_.end(); iter++) {

      std::cout << "  " << (*iter)->get_ast_type_name()
                << " " << (*iter)->get_name() << std::endl;
    }
  }

  // print Threads
  if(threads_.size() != 0) {

    std::cout << " Threads (name: static sensitivity list)" << std::endl;
    for(ThreadVectorIter iter = threads_.begin();
        iter != threads_.end(); iter++) {

      std::cout << "  " << (*iter)->get_name() << ": ";
      (*iter)->print_event_list();
    }
  }

  // print CThreads
  if(cthreads_.size() != 0) {

    std::cout << " CThreads (name, clock)" << std::endl;
    for(CThreadVectorIter iter = cthreads_.begin();
        iter != cthreads_.end(); iter++) {

      std::cout << "  " << (*iter)->get_name() << std::endl;
    }
  }

  // print Methods functions
  if(methods_.size() != 0) {

    std::cout << " Methods (name: static sensitivity list)" << std::endl;
    for(MethodVectorIter iter = methods_.begin();
        iter != methods_.end(); iter++) {

      std::cout << "  " << (*iter)->get_name() << ": ";
      (*iter)->print_event_list();
    }
  }
}


void risc::Module::print_hierarchy(unsigned indent_level)
{
  for(ModuleInstanceVectorIter iter = sub_modules_.begin();
      iter != sub_modules_.end(); iter++) {

    for(unsigned i=0; i<indent_level; i++) {
      std::cout << "  ";
    }

    std::cout << (*iter)->get_ast_type_name()
      << " " << (*iter)->get_name() << std::endl;

    (*iter)->definition_->print_hierarchy(indent_level+1);
  }
}


risc::Event* risc::Module::find_event(SgVariableDefinition* var_def)
{
  for(EventVectorIter iter = events_.begin(); iter != events_.end(); iter++) {
    if(var_def == (*iter)->ast_node_)
      return *iter;
  }
}

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
